import tkinter
from tkinter import ttk
import logging
import jtk
import random

"""
Model of the ALU shown in chapter 4 of Patterson and Hennessy.

 The ALU computes attachment points for the three TextLines and reports them out
 using getLeftX/Y and getRightX/Y. The App uses these points to construct the lines
 and text items, which thus do not belong to the Alu. However, they are placed
 correctly with respect to the body of the Alu and will be placed with it wherever
 its constructor parameters direct. In addition, the App sends references to these
 TextLines to the Alu using setX(), setY(), and setZ(). The Alu methods use these
 references to get and set values in the TextLines.
 
 This means that we can combine the ALU with other devices (viz. registers),
 draw the connecting lines outside all of them, and still not have to
 calculate pixel distances by hand.
"""

class Alu:

    width = 200
    partHeight = 100

    def __init__(self, canvas, x, y, op="CBB"):           ## x,y is top left corner
        self.parent = parent = canvas.master
        self.x, self.y = x, y
        wid = Alu.width
        ph = Alu.partHeight
        totalHeight = 2.5 * ph
        ## calculating some of the y-coordinates is a bit intricate
        tr_y = y+0.75*ph                ## top right corner
        br_y = y+1.75*ph                ## bottom right corner
        bl_y = y+2.5*ph                 ## bottom left corner
        bc_y = (bl_y + br_y) / 2        ## bottom center
        r = canvas.create_polygon(x, y, x+wid, tr_y, x+wid, br_y, x, bl_y,
                            x, bl_y - ph, x + wid/2, y+totalHeight/2,
                            x, y+ph, fill='white', outline='black')
        
        if op == "CBB":
            self.op = ttk.Combobox(parent, postcommand=self.post, width=3, state='readonly', takefocus=False)
            self.op.bind("<<ComboboxSelected>>", self.doOp)
        else:                               # use a Label instead of a Combobox; used by crawlcpu
            self.op = jtk.Label(parent, width=3)
        self.op.set('2')
        jtk.TextLine.ptLen(canvas, (x + wid/2, bc_y), 50, 4, "n", self.op, 40)  # attach Combobox to bottom center
        
        self.cb = jtk.Checkbutton(parent, True, text="#", command=self.checked)
        canvas.create_window(x+totalHeight/2, (tr_y+br_y)/2, window=self.cb, anchor="c", width=35)
        self.nums =  ['0', '1', '2', '6', '7']
        self.signs = ['&', '|', '+', '-', '<']

    def opTrans(self, op):
        if op in self.nums:
            i = self.nums.index(op)
        else:
            i = self.signs.index(op)
        if self.cb.isSelected():
            return self.nums[i]
        else:
            return self.signs[i]

    def getLeftX(self):
        return self.x

    def getRightX(self):
        return self.x + self.width

    def getLeftY(self, position): ## y coordinates of two points down the left edge; 0 --> the top one
        if position == 0:
            return self.y + self.partHeight / 2
        elif position == 1:
            return self.y + 2 * self.partHeight
        else:
            raise ValueError ("left Y-position must be 0 or 1")

    def getRightY(self, position): ## y coordinates of two points down the right edge; 0 --> the top one
        if position == 0:
            return self.y + self.partHeight
        elif position == 1:
            return self.y + 1.5 * self.partHeight
        else:
            raise ValueError ("right Y-position must be 0 or 1")

    def setX(self, x):              ## upper input
        self.xval = x

    def setY(self, y):              ## lower input
        self.yval = y

    def setA(self, a):              ## answer output
        self.aval = a

    def setZ(self, z):              ## zero output
        self.zval = z

## When the Checkbutton is checked the displayed symbol in the Combobox must change
## to the symbol that's in the same position in the other list.
    def checked(self): 
        theOp = self.op.get()
        self.op.set(self.opTrans(theOp))

## This method is called when the Combobox is clicked on. Before this list is
## displayed, this method configures the list to correspond with the Checkbutton
    def post(self):
        if self.cb.isSelected():
            self.op.configure(values=self.nums)
        else:
            self.op.configure(values=self.signs)

    def getValue(self):     ## we separate the calculation from the
        x = self.xval.get() ## posting of the value, which happens in doOp()
        x = int(x)
        y = self.yval.get()
        y = int(y)
        op = self.op.get()
        if op in "0&":     ## bitwise and
            a = x & y
        elif op in "1|":   ## bitwise or
            a = x | y
        elif op in "2+":   ## add
            a = x + y
        elif op in "6-":   ## subtract
            a = x - y
        elif op in "7<":   ## slt (set on less than)
            if x < y: a = 1
            else: a = 0
        logging.debug("answer is {}; str(answer) is '{}'".format(a, str(a)))
        return a

    def doOp(self, evt=None):
        a = self.getValue()
        self.aval.set(str(a))
        if a == 0:                  ## zero output is 1 (True) when answer is 0
            self.zval.set("1")
        else:
            self.zval.set("0")
    
    def resultOK(self, string, dummy):   # callback for validation in jtk.RTVEntry
        if string == '':
            retval = True, '0', 1                 ## third value positions cursor
        else:
            try:
                result = int(string)
            except:
                return False, None                ## no change, so bail out now
            else:
                retval = -1000 <= result <= 1000, self.trim(string)
        self.parent.after(0, self.doOp)
        return retval

    def trim(self, string):  ## get rid of leading zeroes but not entire string
            i = 0
            for i in range(len(string) - 1):
                if string[i] != '0': break
                i += 1
            return string[i:]
               
    def doBad(self):
        self.parent.bell()

if __name__ == "__main__":

    TopLeft_X = 250
    TopLeft_Y = 50
    WIDTH = 700
    HEIGHT = 400

    class App(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            logging.basicConfig(level=logging.INFO)
            win = tkinter.Canvas(self, width=WIDTH, height=HEIGHT)
            win.pack()
            a = Alu(win, TopLeft_X, TopLeft_Y)

            xval = jtk.RTVEntry(self, a.resultOK, a.doBad, width=5, justify="center")
            xval.set(random.randint(-999, 999))
            a.setX(jtk.TextLine.ptLen(win, (a.getLeftX(), a.getLeftY(0)), 100, 20, "e", xval, 50))

            yval = jtk.RTVEntry(self, a.resultOK, a.doBad, width=5, justify="center")
            yval.set(random.randint(-999, 999))
            a.setY(jtk.TextLine.ptLen(win, (a.getLeftX(), a.getLeftY(1)), 100, 20, "e", yval, 50))
        
            aval = jtk.Label(self)
            a.setA(jtk.TextLine.ptLen(win, (a.getRightX(), a.getRightY(1)), 100, 20, "w", aval, 50))
        
            zval = jtk.Label(self)
            a.setZ(jtk.TextLine.ptLen(win, (a.getRightX(), a.getRightY(0)), 100, 1, "w", zval, 50))
        
    def main():
        root = tkinter.Tk()
        root.title("Model ALU")
        a = App(root)
        a.pack()
        root.mainloop()

    main()
