import tkinter
from tkinter import ttk
import logging
import jtk
import random
import registers
import alu

"""
Model the CPU depicted in chapter 4 of Patterson and Hennessy.
Only do R-format instructions.
"""

class Cpu:
    def __init__(self, canvas, x, y, sep):
        parent = canvas.master
        self.r = r = registers.RegisterFile(canvas, x, y)
        r.cb.set(True)
        r.b.configure(command=self.doButton)
        
        cbvalues = r.getComboValues()

        r1op = ttk.Combobox(parent, values=cbvalues, width=3,    ## read reg 1
                                    state='readonly', takefocus=False)
        r1op.bind("<<ComboboxSelected>>", self.doRD1)
        r1op.set(random.randint(0, len(cbvalues) - 1))
        r.setRR1(jtk.TextLine.ptLen(canvas, (r.getLeftX(), r.getLeftY(0)), 100, 10, 'e', r1op, 45))
        
        r2op = ttk.Combobox(parent, values=cbvalues, width=3,    ## read reg 2
                                    state='readonly', takefocus=False)
        r2op.bind("<<ComboboxSelected>>", self.doRD2)
        r2op.set(random.randint(0, len(cbvalues) - 1))
        r.setRR2(jtk.TextLine.ptLen(canvas, (r.getLeftX(), r.getLeftY(1)), 100, 10, 'e', r2op, 45))
        
        wop = ttk.Combobox(parent, values=cbvalues, width=3,     ## write reg
                                    state='readonly', takefocus=False)
        wop.set(random.randint(0, len(cbvalues) - 1))
        r.setWR(jtk.TextLine.ptLen(canvas, (r.getLeftX(), r.getLeftY(2)), 100, 10, "e", wop, 45))
        
        self.a = a = alu.Alu(canvas, x+sep, y)
        ## here we need to make a TextLine connecting the rd1 output of the Register
        ## File to the x input of the alu. Because the Label goes in the center of the
        ## line, we have to use two endpoints to construct the line. The left endpoint
        ## will have x coordinate the same as the right side of the register file and
        ## y coordinate to match the x input of the alu. The alu's x input is the right
        ## endpoint. The TextLine we construct becomes both the RD1 output of the
        ## RegisterFile and the x input of the Alu
        regpt = (r.getRightX(), a.getLeftY(0))
        alupt = (a.getLeftX(), a.getLeftY(0))
        self.rd1val = jtk.Label(parent)                             ## read data 1
        rdl = jtk.TextLine.endPoints(canvas, regpt, alupt, 20, "c", self.rd1val, 40)
        r.setRD1(rdl)
        a.setX(rdl)

        ## similar comments for the TextLine connecting RD2 and the y input of the Alu.
        regpt = (r.getRightX(), a.getLeftY(1))
        alupt = (a.getLeftX(), a.getLeftY(1))
        self.rd2val = jtk.Label(canvas)                             ## read data 2
        rdl = jtk.TextLine.endPoints(canvas, regpt, alupt, 20, "c", self.rd2val, 40)
        r.setRD2(rdl)
        a.setY(rdl)

        ## The line connecting Alu output to RegisterFile WD input is in five pieces.
        ## Only the bottom portion has a Label, and that Label is both the Alu Z output
        ## and the RegisterFile WD input. The other four pieces are just lines. We use
        ## TextLines instead of canvas.create_line more for practice than anything else.
        ## Would likely have been simpler the other way.

        ## We start by computing the coordinates of the four intersection points of the
        ## five lines: topLeft, bottomLeft, topRight, and bottomRight 
        stub = 100                                       ## length of each of the top two horizontal lines
        regX, regY = r.getLeftX(), r.getLeftY(3)         ## RegisterFile's WriteData input point
        topLeft = regX-stub, regY                        ## regX, regY is the right end of the stub line
        bottomLeft = topLeft[0], regY+2*stub             ## same x, move y down
        aluX, aluY = (a.getRightX(), a.getRightY(1))     ## Alu answer output point
        topRight = aluX+stub, aluY                       ## right end of stub line
        bottomRight = topRight[0], bottomLeft[1]
        jtk.TextLine.ptLen(canvas, topLeft, stub, 20, 'w', None, None)              ## left horiz            
        jtk.TextLine.endPoints(canvas, topLeft, bottomLeft, 20, 's', None, None)    ## left vert
        jtk.TextLine.ptLen(canvas, topRight, stub, 20, 'e', None, None)             ## right horiz                
        jtk.TextLine.endPoints(canvas, topRight, bottomRight, 20, 's', None, None)  ## right vert

        ## The bottom line has the Label, which is both Alu's answer output and RegisterFile's WD input.
        self.wdval = jtk.Label(parent)
        wdl = jtk.TextLine.endPoints(canvas, bottomLeft, bottomRight, 20, 'c', self.wdval, 40)
        a.setA(wdl)
        r.setWD(wdl)
        
        self.zval = jtk.Label(parent)                               ## the Alu's zero output bit
        a.setZ(jtk.TextLine.ptLen(canvas, (a.getRightX(), a.getRightY(0)), stub, 1, 'w', self.zval, 15))
        
        self.r.doRD1()      ## can't use self.RD1 here because it calls a.doOp,
        self.r.doRD2()      ## which needs RD2 to be called before it can work
        self.a.doOp()
        a.cb.set(False)
        a.checked()         ## changing Checkbutton from program does not call its command method, so we have to

    def doRD1(self, evt=None):
        self.r.doRD1()                          ## set RegisterFile output value
        self.a.doOp()                           ## and set new Alu output value

    def doRD2(self, evt=None):
        self.r.doRD2()
        self.a.doOp()
        
    def doButton(self, evt=None):               ## write new register value
        self.r.doButton()
        self.doRD1()                            ## perhaps we wrote into
        self.doRD2()                            ## a reading register

if __name__ == "__main__":
    TopLeft_X = 200
    TopLeft_Y = 100
    Separation = 400

    class App(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            logging.basicConfig(level=logging.INFO)
            self.canvas = canvas = tkinter.Canvas(self, width=1000, height=700)
            canvas.pack()
        
            Cpu(canvas, TopLeft_X, TopLeft_Y, Separation)

    def main():
        root = tkinter.Tk()
        root.title("Model CPU for R-format Instructions")
        a = App(root)
        a.pack()
        root.mainloop()
        
    main()
