"""
The program is composed of a model (this file) and a number of
views on the model. Think of the model as an odometer.
It holds a long value and a modulus (turnover point).
All computer arithmetic has a modulus, just as an odometer does.
The views display the value in traditional positional notation using
bases in the range 2-36. Any view may show signed or unsigned numerals.
The model may have no value at all, in which case the views will be blank.
Unlike an odometer the Model can change its modulus at any time.
Because of this the Model also maintains a MAXMODULUS value. The Modulus
object maintains 2 <= modulus <= MAXMODULUS as an invariant.

This implementation of Model is capable of displaying its value and modulus
in Labels that are updated whenever they change. Use of this capability
is optional: if the parameter parent has a value of None, then this
visibility option is disabled.
"""
import tkinter
from tkinter import ttk
import logging
import math

class Model(ttk.Frame):
    def __init__(self, parent=None, maxModulus=1<<16):
        super().__init__(parent)
        self.parent = parent
        self.__value = None
        self.__modulus = None
        self.MAXMODULUS = maxModulus
        if parent is not None:
            valLabLab = ttk.Label(self, text="model value:").grid(
                                row=0, column=0, padx=20)
            self.valLab = ttk.Label(self)
            self.valLab.grid(row=0, column=1)
            modLabLab = ttk.Label(self, text="modulus:").grid(
                                row=1, column=0, padx=20)
            self.modLab = ttk.Label(self)
            self.modLab.grid(row=1, column=1)
            self.setValue(None)
            self.setModulus(maxModulus)

    def getValue(self):
        assert self.__value is not None , "called getValue() when there is no value"
        logging.info("model returning {}".format(self.__value))
        return self.__value
    
    def isGoodValue(self):
        return self.__value is not None

    def setValue(self, newValue):
        assert newValue is None or 0 <= newValue < self.__modulus
        self.__value = newValue
        logging.info("model accepting {}".format(newValue))
        if self.parent is not None:
            self.valLab["text"] = str(newValue)

    def getModulus(self):
        return self.__modulus

    def setModulus(self, newModulus):
        assert 2 <= newModulus <= self.MAXMODULUS
        if self.isGoodValue():
            v = self.getValue()
            ## if some view is showing a negative value
            if v > self.maxPossible(True) and self.master.someSigned():
                v = self.getModulus() - v   ## abs(v)
                wasNegative = True
            else:
                wasNegative = False
            v = v % newModulus              ## make it fit        
            if wasNegative:
                v = newModulus - v            
        self.__modulus = newModulus
        if self.parent is not None:
            self.modLab["text"] = str(newModulus)            
        if self.isGoodValue():
            self.setValue(v)
            self.master.updateDigits()
            self.master.updateViews(None)

    def maxPossible(self, isSigned):
        modulus = self.getModulus()
        if isSigned:
            m = math.floor((modulus-1)/2)  ## works for even or odd modulus
        else:
            m = modulus - 1
        logging.debug("maxPossible sez {}".format(m))
        return m

    def minPossible(self, isSigned):
        modulus = self.getModulus()
        if isSigned:
            m = -math.floor(modulus/2)  ## works for even or odd modulus
        else:
            m = 0
        logging.debug("minPossible sez {}".format(m))
        return m
    
class App(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent)
        self.m = m = Model(self, 1<<16)         ## pick one
##        self.m = m = Model(None, 1<<16)        ## of these
        m.pack()
        b = ttk.Button(self, text="next test", command=self.test)
        b.pack(padx=75)
        self.i = 0
        
    def test(self):
        name = 'test' + str(self.i)
        try:
            code = self.__getattribute__(name)
        except AttributeError:  ## went off the end, so wrap
            self.i = 0          ## back around to the beginning
            code = self.__getattribute__('test0')
        print("\nrunning test {}:".format(self.i))
        self.i += 1
        code() ## NOT code(self)
        
    def test0(self):
        print("model sez isGoodValue is {}".format(self.m.isGoodValue()))
        print("setting model value to {}".format(3))
        self.m.setValue(3)
        print("model sez value is {} and isGoodValue is {}"
              .format(self.m.getValue(), self.m.isGoodValue()))
    def test1(self):
        print("setting modulus to {}".format(8))
        self.m.setModulus(8)
        print("getModulus gives us {}".format(self.m.getModulus()))
        print("maxPossible sez {} and {}"
              .format(self.m.maxPossible(True),self.m.maxPossible(False)))
        print("minPossible sez {} and {}"
              .format(self.m.minPossible(True),self.m.minPossible(False)))
    def test2(self):
        print("setting to no value")
        self.m.setValue(None)
        print("model sez isGoodValue is {}".format(self.m.isGoodValue()))
    def test3(self):
        print("calling getValue() anyway")
        print(self.m.getValue())
    def test4(self):
        print("trying direct access to __value")
        print(self.__value)
    def test5(self):
        print("MAXMODULUS is {}".format(self.m.MAXMODULUS))
        self.m.setModulus(self.m.MAXMODULUS)
    def test6(self):
        print("setting modulus to {}".format(9))
        self.m.setModulus(9)
        print("getModulus gives us {}".format(self.m.getModulus()))
        print("maxPossible sez {} and {}"
              .format(self.m.maxPossible(True),self.m.maxPossible(False)))
        print("minPossible sez {} and {}"
              .format(self.m.minPossible(True),self.m.minPossible(False)))

    def updateViews(self, skip):
        pass

    def updateDigits(self):
        pass

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    root = tkinter.Tk()
    root.title("testing Model")
    a = App(root)
    a.pack()
    root.mainloop()

