import tkinter
from tkinter import ttk
import logging
import jtk
import random
import blinker


"""
Model the register file depicted in chapter 4 of Patterson and Hennessy.
There are 8 registers, we use Comboboxes to pick register numbers, and
we use a jtk.RTVEntry to type in the value for the data bus. These external
items are not part of the RegisterFile so that it can be used with other
components (e.g. ALU) to form a larger system (e.g. CPU). In this latter
case, the external components will be different.

I thin I can rewrite it so that the RegisterFile class creates all the
external lines and data areas, but only when instructed to do so by
the containing application. This way, when the RegisterFile is part of
a larger component, the containing application will create the external
parts itself instead of asking this class to do it.
"""

class Register:
    def __init__(self, canvas, x, y, i):
        self.value = random.randint(-10, 100)
        self.canvas = canvas
        canvas.create_rectangle(x, y+10, x+50, y+40, fill='white')
        self.labelValue = str(i)+" :"
        self.label = canvas.create_text(x-5, y+25, anchor="e", text=self.labelValue)
        self.display = canvas.create_text(x+25, y+25, anchor="c", text=self.value)
        self.bl = blinker.Blinker((canvas, self.label))                             # not used in this App
        self.bd = blinker.Blinker((canvas, self.display))                           # ditto

    def get(self):
        return self.value

    def showValue(self, where, value):
        self.canvas.dchars(where, 0, 'end')
        self.canvas.insert(where, 0, value)

    def set(self, newValue):
        self.value = newValue
        self.showValue(self.display, newValue)

    def setLabel(self, newValue):
        self.labelValue = newValue
        self.showValue(self.label, newValue)


            
        

class RegisterFile:
    def __init__(self, canvas, x, y, drawButton=True, buttonCallback=None):    ## (x, y) = top-left corner
        width, height = 200, 400                               ## dimensions of containing box
        COUNT = 8                                             ## number of registers
        self.parent = parent = canvas.master
        r = canvas.create_rectangle(x, y, x+width, y+height, fill='white', outline='black')
        self.regs = []                      ## the registers
        self.cbs = []                       ## the register numbers for the Comboboxes
        for i in range(COUNT):
            self.regs.append(Register(canvas, x+75, y+50*i, i))
            self.cbs.append(str(i))
        if drawButton:                  ## allow for no Button
            self.b = b = ttk.Button(parent, text="write", command=self.doButton)
            jtk.TextLine.ptLen(canvas, (x + width/2, y+height), 50, 4, "n", b, 75)
        self.cb = jtk.Checkbutton(parent, True, text="show", command=self.checkboxChanged)
        canvas.create_window(x+width-40, y+height/2, window=self.cb, width=60)
        
        self.x, self.y, self.width, self.height = x, y, width, height
        self.rightYstart = self.height / 6
        self.rightYincrement = self.height / 3
        self.buttonCallback = buttonCallback

        
    def get(self, i):                   ## return value of register i
        return self.regs[i].get()

    def set(self, i, newValue):         ## set value of register i
        self.regs[i].set(newValue)
        self.doRD1()                    ## What we just did may have changed the value in
        self.doRD2()                    ## one of the output registers. Make sure they are correct.

    def getComboValues(self):           ## the list of values displayed in the Comboboxes
        return self.cbs

    ## The terminology in the next six methods is from Patterson and Henessey.
    ## Each of these objects being set is a line with a window attached to it.
    ## The RR1, RR2, and WR lines have Comboboxes that can choose any register number.
    ## WD has an RTVEntry that can receive a typed value.
    ## RD1 and RD2 have Labels that can display the values in the two registers selected by RR1 and RR2.

    def setRR1(self, line):            ## input line that chooses Read Register 1
        self.rr1Input = line

    def setRR2(self, line):            ## input line that chooses Read Register 2
        self.rr2Input = line

    def setWR(self, line):             ## input line that chooses the Write Register
        self.wrInput = line

    def setWD(self, line):             ## Write Data input line
        self.wdInput = line

    def setRD1(self, line):            ## Read Data 1 output line
        self.rd1Output = line

    def setRD2(self, line):            ## Read Data 2 output
        self.rd2Output = line

    def checkboxChanged(self):                      ## called whenever the Checkbox is clicked
        willBeVisible = self.cb.isSelected()
        self.rd1Output.setVisible(willBeVisible)
        self.rd2Output.setVisible(willBeVisible)

    ## These four methods return the coordinates of the attachment points around the edge of the rectangle.

    def getLeftX(self):                ## x coordinate of the left edge
        return self.x

    def getRightX(self):               ## x coordinate of the right edge
        return self.x + self.width

    def getLeftY(self, position):     ## y coordinates of four points down the left edge; 0 --> the top one
        assert 0 <= position <= 3
        return self.y + ((2 * position + 1) / 8) * self.height

    def getRightY(self, position):    ## y coordinates of two points down the right edge; 0 --> the top one
        assert 0 <= position <= 1
        return self.y + self.rightYstart + position * self.rightYincrement

    ## The RTVEntry is a jtk.RTVEntry, which does validation on a keysstroke-by-keysstroke basis.
    ## The next three methods are called by the jtk.RTVEntry in the process of doing that.
   
    def resultOK(self, string, dummy):           ## validation routine for the RTVEntry
        if string == '':
            retval = True, '0', 1                 ## third value positions cursor
        else:
            try:
                result = int(string)
            except:
                return False, None                ## no change, so bail out now
            else:
                retval = -100 <= result <= 100, self.trim(string)
        return retval

    def trim(self, string):                      ## get rid of leading zeroes but not entire string
            i = 0
            for i in range(len(string) - 1):
                if string[i] != '0': break
                i += 1
            return string[i:]
        
    def doBad(self):                             ## called when vallidation fails
        self.parent.bell()

    def doButton(self, evt=None):               ## called when the "write" button is pushed
        reg = self.wrInput.get()                  ## find out which register to write to
        reg = int(reg)
        val = self.wdInput.get()                  ## get the value to put there
        logging.debug("changing value in register {} to {}".format(reg, val))
        self.set(reg, val)                      ## copy it into the correct register
        if self.buttonCallback is not None:
            self.buttonCallback()

    ## This next method is called from three places: self.set(,) above,
    ## the Combobox in the rr1Input line (as specified in the constructor for RegisterApp below),
    ## and from that same constructor itself.
    ## It's job is to make sure that rd1Output correctly reflects the contents of the register.
    ## If rr1Input or rd1Output does not exist; catch the exception and pass
        
    def doRD1(self, evt=None):
        try:
            newVal = self.RDvalue(self.rr1Input)       ## get correct value from correct register
            self.rd1Output.set(newVal)                  ## and copy it to the rd1Output
        except AttributeError as e:
            pass
 

## same, but for rr2Input
    def doRD2(self, evt=None):
        try:
            newVal = self.RDvalue(self.rr2Input)
            self.rd2Output.set(newVal)
        except AttributeError as e:
            pass

    def RDvalue(self, rrInputLine):                 ## rrInputLine is rr1Input or rr2Input                         
        reg = rrInputLine.get()                     ## look up register number from input line
        reg = int(reg)
        return self.get(reg)                        ## and get the value in that register

if __name__ == "__main__":
    TopLeft_X = 200
    TopLeft_Y = 100

    ## Here's a demonstration of how to use a RegisterFile object.
    ## First we create the registerFile object and then we create the
    ## lines connecting to it. In this case the lines are TextLine objects.
    ## Each line is placed at coordinates retrieved from the RegisterFile object.
    ## For each line we get the coordinates of the point on it that attaches
    ## to the RegisterFile object's outline rectangle. The other end of the
    ## line is computed by the line itself from its length and the direction
    ## in which it runs. The direction is computed from the parameter ('e' or 'w')
    ## that specifies which side of the input/output object the line attaches to.
    ## The local variables outputObject and inputObject are "recycled" since each
    ## has no further use once it has been used to initialize a TextLine ovbect.


    class RegisterApp(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            logging.basicConfig(level=logging.INFO)
            self.win = win = tkinter.Canvas(self, width=600, height=700)
            win.pack()
            
            r = RegisterFile(win, TopLeft_X, TopLeft_Y)                 ## create RegisterFile objack
            cbvalues = r.getComboValues()                         ## list of labels for the Comboboxes
            
            
            outputObject = jtk.Label(win)                                 ## read data 1
            r.setRD1(jtk.TextLine.ptLen(win, (r.getRightX(), r.getRightY(0)), 100, 35, "w", outputObject, 45))
            
            outputObject = jtk.Label(win)                                 ## read data 2
            r.setRD2(jtk.TextLine.ptLen(win, (r.getRightX(), r.getRightY(1)), 100, 35, "w", outputObject, 45))

            inputObject = ttk.Combobox(self, values=cbvalues, width=3,    ## read reg 1
                                        state='readonly', takefocus=False)
            inputObject.bind("<<ComboboxSelected>>", r.doRD1)   ## show new value when register changes
            inputObject.set(random.randint(0, len(cbvalues) - 1))
            r.setRR1(jtk.TextLine.ptLen(win, (r.getLeftX(), r.getLeftY(0)), 100, 10, 'e', inputObject, 45))
            r.doRD1()                               ## make it start out showing the right value
            
            inputObject = ttk.Combobox(self, values=cbvalues, width=3,    ## read reg 2
                                        state='readonly', takefocus=False)
            inputObject.bind("<<ComboboxSelected>>", r.doRD2)
            inputObject.set(random.randint(0, len(cbvalues) - 1))
            r.setRR2(jtk.TextLine.ptLen(win, (r.getLeftX(), r.getLeftY(1)), 100, 10, 'e', inputObject, 45))
            r.doRD2()      
            
            inputObject = ttk.Combobox(self, values=cbvalues, width=3,     ## write reg
                                        state='readonly', takefocus=False)
            inputObject.set(random.randint(0, len(cbvalues) - 1))
            r.setWR(jtk.TextLine.ptLen(win, (r.getLeftX(), r.getLeftY(2)), 100, 10, "e", inputObject, 45))
            
    ## this is last because in some use cases validation is premature before rd1Output and rd2Output are set
            inputObject = jtk.RTVEntry(self, r.resultOK, r.doBad, width=20, justify="center")
            inputObject.set(random.randint(-10, 100))
            r.setWD(jtk.TextLine.ptLen(win, (r.getLeftX(), r.getLeftY(3)), 100, 35, "e", inputObject, 60))
            

    ## end of class App

    def main() -> object:
        root = tkinter.Tk()
        root.title("Model Register File")
        ra = RegisterApp(root)
        ra.pack()
        root.mainloop()
        
    main()
